# Random Instance Generator

Random instance generator used in the papers:

* [The multiperiod two-dimensional non-guillotine cutting stock problem with usable leftovers](https://onlinelibrary.wiley.com/doi/abs/10.1111/itor.12648)
* [A forward-looking matheuristic approach for the multi-period two-dimensional non-guillotine cutting stock problem with usable leftovers](https://www.ime.usp.br/~egbirgin/)

## Compile
* `cmake CMakeLists.txt`
* `make`

## Run

```
./InstanceGenerator -s <s> -p1 <p1> -p2 <p2> -p3 <p3> -p4 <p4> -p5 <p5> -p6 <p6> -p7 <p7> -p8 <p8> -p9 <p9> -p10 <p10> -p11 <p11> -p12 <p12>
```

where:

* `<s>`: value to be used as seed by the pseudo-random number generator algorithm;
* `<p1>`: minimum number of period;
* `<p2>`: maximum number of period;
* `<p3>`: minimum number of object type;
* `<p4>`: maximum number of object type;
* `<p5>`: lower bound of objects size;
* `<p6>`: upper bound of objects size;
* `<p7>`: minimum number of item type;
* `<p8>`: maximum number of item type;
* `<p9>`: minimum number of item per type;
* `<p10>`: maximum number of item per type;
* `<p11>`: lower bound of items size;
* `<p12>`: upper bound of items size;

## Exemple of use:

```
./InstanceGenerator -s 2 -p1 4 -p2 4 -p3 2 -p4 3 -p5 30 -p6 100 -p7 2 -p8 5 -p9 1 -p10 3 -p11 5 -p12 20
```

Possible result:

```
4

3	8	10	9	
11	10	19	
8	18	10	
19	19	19	18	7	7	9	18	
13	13	13	7	6	6	6	18	
17	17	11	11	10	10	10	18	18	18	
10	10	5	5	16	16	16	7	7	7	
7	7	7	20	20	20	14	14	14	
19	19	19	19	19	19	7	7	7	


2
7	11	
6	5	

3	3	3	3	
59	79	33	
86	71	40	
1	1	1	
69	85	62	
54	43	69	
1	1	1	
59	59	96	
32	37	43	
1	1	1	
46	39	39	
54	64	64	
1	1	1	
```

The first value represent the number of periods ($P$). The next line contains $P$ values indicating the number of items ($n_s$, $s = 1, \dots, P$). The next $2 \times P$ lines contains the dimensions of each item for each period, following the pattern:

$w_{11}~~w_{12}~~\dots~~w_{1n_1}$
$h_{11}~~h_{12}~~\dots~~h_{1n_1}$
$w_{21}~~w_{22}~~\dots~~w_{2n_2}$
$h_{21}~~h_{22}~~\dots~~h_{2n_2}$
$\vdots$
$w_{P1}~~w_{P2}~~\dots~~w_{Pn_P}$
$h_{P1}~~h_{P2}~~\dots~~h_{Pn_P}$

The next line contains the amount of catalog itens ($d$). The next two lines contains the dimensions of each item in the catalog, following the pattern:
$\bar{w}_1~~\bar{w}_1~~\dots~~\bar{w}_d$
$\bar{h}_1~~\bar{h}_1~~\dots~~\bar{h}_d$

Next, the amount of available objects ($m_s$, $s = 1, \dots, P$) in each period is shown, following the pattern:

$W_{11}~~W_{12}~~\dots~~W_{1m_1}$
$H_{11}~~H_{12}~~\dots~~H_{1m_1}$
$c_{11}~~c_{12}~~\dots~~c_{1m_1}$
$W_{21}~~W_{22}~~\dots~~W_{2m_2}$
$H_{21}~~H_{22}~~\dots~~H_{2m_2}$
$c_{21}~~c_{22}~~\dots~~c_{2m_2}$
$\vdots$
$W_{P1}~~W_{P2}~~\dots~~W_{Pm_P}$
$H_{P1}~~H_{P2}~~\dots~~H_{Pm_P}$
$c_{P1}~~c_{P2}~~\dots~~c_{Pm_P}$

In summary, the instance data generated by the random instance generator follows the pattern:

$P$

$n_1~~n_2~~\dots~~n_P$
$w_{11}~~w_{12}~~\dots~~w_{1n_1}$
$h_{11}~~h_{12}~~\dots~~h_{1n_1}$
$w_{21}~~w_{22}~~\dots~~w_{2n_2}$
$h_{21}~~h_{22}~~\dots~~h_{2n_2}$
$\vdots$
$w_{P1}~~w_{P2}~~\dots~~w_{Pn_P}$
$h_{P1}~~h_{P2}~~\dots~~h_{Pn_P}$

$d$
$\bar{w}_1~~\bar{w}_1~~\dots~~\bar{w}_d$
$\bar{h}_1~~\bar{h}_1~~\dots~~\bar{h}_d$

$m_1~~m_2~~\dots~~m_P$
$W_{11}~~W_{12}~~\dots~~W_{1m_1}$
$H_{11}~~H_{12}~~\dots~~H_{1m_1}$
$c_{11}~~c_{12}~~\dots~~c_{1m_1}$
$W_{21}~~W_{22}~~\dots~~W_{2m_2}$
$H_{21}~~H_{22}~~\dots~~H_{2m_2}$
$c_{21}~~c_{22}~~\dots~~c_{2m_2}$
$\vdots$
$W_{P1}~~W_{P2}~~\dots~~W_{Pm_P}$
$H_{P1}~~H_{P2}~~\dots~~H_{Pm_P}$
$c_{P1}~~c_{P2}~~\dots~~c_{Pm_P}$